// • ▌ ▄ ·.  ▄▄▄·  ▄▄ • ▪   ▄▄· ▄▄▄▄·  ▄▄▄·  ▐▄▄▄  ▄▄▄ .
// ·██ ▐███▪▐█ ▀█ ▐█ ▀ ▪██ ▐█ ▌▪▐█ ▀█▪▐█ ▀█ •█▌ ▐█▐▌·
// ▐█ ▌▐▌▐█·▄█▀▀█ ▄█ ▀█▄▐█·██ ▄▄▐█▀▀█▄▄█▀▀█ ▐█▐ ▐▌▐▀▀▀
// ██ ██▌▐█▌▐█ ▪▐▌▐█▄▪▐█▐█▌▐███▌██▄▪▐█▐█ ▪▐▌██▐ █▌▐█▄▄▌
// ▀▀  █▪▀▀▀ ▀  ▀ ·▀▀▀▀ ▀▀▀·▀▀▀ ·▀▀▀▀  ▀  ▀ ▀▀  █▪ ▀▀▀
//      Magicbane Emulator Project © 2013 - 2022
//                www.magicbane.com


package engine.objects;

import engine.InterestManagement.WorldGrid;
import engine.gameManager.BuildingManager;
import engine.gameManager.ChatManager;
import engine.gameManager.DbManager;
import engine.gameManager.ZoneManager;
import engine.mbEnums;
import engine.net.ByteBufferWriter;
import engine.net.client.msg.ErrorPopupMsg;
import engine.server.MBServerStatics;
import org.pmw.tinylog.Logger;

import java.net.UnknownHostException;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.concurrent.ConcurrentHashMap;

import static engine.gameManager.DbManager.MineQueries;
import static engine.gameManager.DbManager.getObject;
import static engine.math.FastMath.sqr;

public class Mine extends AbstractGameObject {

    public static ConcurrentHashMap<Mine, Integer> mineMap = new ConcurrentHashMap<>(MBServerStatics.CHM_INIT_CAP, MBServerStatics.CHM_LOAD, MBServerStatics.CHM_THREAD_LOW);
    public static ConcurrentHashMap<Integer, Mine> towerMap = new ConcurrentHashMap<>(MBServerStatics.CHM_INIT_CAP, MBServerStatics.CHM_LOAD, MBServerStatics.CHM_THREAD_LOW);
    private final String zoneName;
    private final Zone parentZone;
    public boolean isActive = false;
    public PlayerCharacter lastClaimer;
    public boolean wasClaimed = false;
    // Not persisted to DB
    public String guildName;
    public GuildTag guildTag;
    public String nationName;
    public GuildTag nationTag;
    private mbEnums.ResourceType production;
    private Guild owningGuild;
    private int flags;
    private int buildingID;
    private MineProduction mineType;

    /**
     * ResultSet Constructor
     */
    public Mine(ResultSet rs) throws SQLException, UnknownHostException {
        super(rs);

        this.mineType = MineProduction.getByName(rs.getString("mine_type"));

        int ownerUID = rs.getInt("mine_ownerUID");
        this.buildingID = rs.getInt("mine_buildingUID");
        this.flags = rs.getInt("flags");
        int parent = rs.getInt("parent");
        this.parentZone = ZoneManager.getZoneByUUID(parent);
        this.zoneName = this.parentZone.parent.zoneName;

        this.owningGuild = Guild.getGuild(ownerUID);
        Guild nation = null;

        if (this.owningGuild.isEmptyGuild()) {
            this.guildName = "";
            this.guildTag = GuildTag.ERRANT;
            nation = Guild.getErrantGuild();
            this.owningGuild = Guild.getErrantGuild();
        } else {
            this.guildName = this.owningGuild.getName();
            this.guildTag = this.owningGuild.getGuildTag();
            nation = this.owningGuild.getNation();
        }

        if (!nation.isEmptyGuild()) {
            this.nationName = nation.getName();
            this.nationTag = nation.getGuildTag();
        } else {
            this.nationName = "";
            this.nationTag = GuildTag.ERRANT;
        }

        this.production = mbEnums.ResourceType.valueOf(rs.getString("mine_resource"));
        this.lastClaimer = null;

    }

    public static void releaseMineClaims(PlayerCharacter playerCharacter) {

        if (playerCharacter == null)
            return;

        for (Mine mine : Mine.getMines()) {

            if (mine.lastClaimer != null)
                if (mine.lastClaimer.equals(playerCharacter)) {
                    mine.lastClaimer = null;
                    mine.updateGuildOwner(null);
                }

        }
    }

    public static void SendMineAttackMessage(Building mine) {

        if (mine.getBlueprint() == null)
            return;

        if (mine.getBlueprint().getBuildingGroup() != mbEnums.BuildingGroup.MINE)
            return;


        if (mine.getGuild().isEmptyGuild())
            return;

        if (mine.getGuild().getNation().isEmptyGuild())
            return;

        if (mine.getTimeStamp("MineAttack") > System.currentTimeMillis())
            return;

        mine.getTimestamps().put("MineAttack", System.currentTimeMillis() + MBServerStatics.ONE_MINUTE);

        ChatManager.chatNationInfo(mine.getGuild().getNation(), mine.getName() + " in " + mine.getParentZone().parent.zoneName + " is Under attack!");
    }

    public static void loadAllMines() {

        try {

            //Load mine resources
            MineProduction.addResources();

            //pre-load all building sets
            ArrayList<Mine> serverMines = MineQueries.GET_ALL_MINES_FOR_SERVER();

            for (Mine mine : serverMines) {
                Mine.mineMap.put(mine, mine.buildingID);
                Mine.towerMap.put(mine.buildingID, mine);
            }

        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    /*
     * Getters
     */

    public static Mine getMineFromTower(int towerID) {
        return Mine.towerMap.get(towerID);
    }

    public static void serializeForClientMsg(Mine mine, ByteBufferWriter writer) {
        writer.putInt(mine.getObjectType().ordinal());
        writer.putInt(mine.getObjectUUID());
        writer.putInt(mine.getObjectUUID()); //actually a hash of mine
        writer.putString(mine.mineType.name);
        writer.putString(mine.zoneName);
        writer.putInt(mine.production.resourceHash);
        writer.putInt(mine.production.mine_production);
        writer.putInt(mine.getModifiedProductionAmount()); //TODO calculate range penalty here
        writer.putInt(3600); //window in seconds

        // Errant mines are currently open.  Set time to now.

        LocalDateTime mineOpenTime = LocalDateTime.now().withMinute(0).withSecond(0).withNano(0);

        // Mine times are those of the nation not individual guild.

        Guild mineNatonGuild = mine.getOwningGuild().getNation();

        // Adjust the serialized mine time based upon whether
        // the Guild's mine window has passed or not and if it was claimed.
        // If a mine is active serialize current datetime irrespective
        // of any claim.

        if (mineNatonGuild.isEmptyGuild() == false && mine.isActive == false) {

            int guildWOO = mineNatonGuild.getNation().getMineTime();
            LocalDateTime guildMineTime = mineOpenTime.withHour(guildWOO);

            if (mineOpenTime.isAfter(guildMineTime) || mine.wasClaimed == true)
                mineOpenTime = guildMineTime.plusDays(1);
            else
                mineOpenTime = guildMineTime;

        }

        writer.putLocalDateTime(mineOpenTime);
        writer.putLocalDateTime(mineOpenTime.plusHours(1));
        writer.put(mine.isActive ? (byte) 0x01 : (byte) 0x00);

        Building mineTower = BuildingManager.getBuilding(mine.buildingID);
        writer.putFloat(mineTower.getLoc().x);
        writer.putFloat(mineTower.getParentZone().getLoc().y);
        writer.putFloat(mineTower.getLoc().z);

        writer.putInt(mine.isExpansion() ? mine.mineType.xpacHash : mine.mineType.hash);

        writer.putString(mine.guildName);
        GuildTag._serializeForDisplay(mine.guildTag, writer);
        writer.putString(mine.nationName);
        GuildTag._serializeForDisplay(mine.nationTag, writer);
    }

    public static ArrayList<Mine> getMinesForGuild(int guildID) {

        ArrayList<Mine> mineList = new ArrayList<>();

        // Only inactive mines are returned.

        for (Mine mine : Mine.mineMap.keySet()) {
            if (mine.owningGuild.getObjectUUID() == guildID)
                mineList.add(mine);
        }
        return mineList;
    }

    /*
     * Database
     */
    public static Mine getMine(int UID) {
        return MineQueries.GET_MINE(UID);

    }

    public static ArrayList<Mine> getMines() {
        return new ArrayList<>(mineMap.keySet());
    }

    public static boolean validateClaimer(PlayerCharacter playerCharacter) {

        // Method validates that the claimer meets
        // all the requirements to claim; landed
        // guild with a warehouse, etc.

        Guild playerGuild;

        //verify the player exists

        if (playerCharacter == null)
            return false;

        //verify the player is in valid guild

        playerGuild = playerCharacter.getGuild();

        // Can't claim something if you don't have a guild!

        if (playerGuild.isEmptyGuild())
            return false;

        if (playerGuild.getNation().isEmptyGuild())
            return false;

        // Guild must own a city to hold a mine.

        City guildCity = playerGuild.getOwnedCity();

        if (guildCity == null)
            return false;

        if (guildCity.warehouse == null) {
            ErrorPopupMsg.sendErrorMsg(playerCharacter, "No Warehouse exists for this claim.");
            return false;
        }

        // Number of mines is based on the rank of the nation's tree.

        City nationCapitol = playerGuild.getNation().getOwnedCity();

        Building nationCapitolTOL = nationCapitol.getTOL();

        if (nationCapitolTOL == null)
            return false;

        int treeRank = nationCapitolTOL.getRank();

        if (treeRank < 1)
            return false;

        if (guildUnderMineLimit(playerGuild.getNation(), treeRank) == false) {
            ErrorPopupMsg.sendErrorMsg(playerCharacter, "Your nation cannot support another mine.");
            return false;
        }

        return true;
    }

    private static boolean guildUnderMineLimit(Guild playerGuild, int tolRank) {

        int mineCnt = 0;

        mineCnt += Mine.getMinesForGuild(playerGuild.getObjectUUID()).size();

        for (Guild guild : playerGuild.getSubGuildList())
            mineCnt += Mine.getMinesForGuild(guild.getObjectUUID()).size();

        return mineCnt <= tolRank;
    }

    public boolean changeProductionType(mbEnums.ResourceType resource) {
        if (!this.validForMine(resource))
            return false;
        //update resource in database;
        if (!MineQueries.CHANGE_RESOURCE(this, resource))
            return false;

        this.production = resource;
        return true;
    }

    public MineProduction getMineType() {
        return this.mineType;
    }

    public void setMineType(String type) {
        this.mineType = MineProduction.getByName(type);
    }

    public String getZoneName() {
        return this.zoneName;
    }

    public mbEnums.ResourceType getProduction() {
        return this.production;
    }

    public boolean getIsActive() {
        return this.isActive;
    }

    public Guild getOwningGuild() {
        if (this.owningGuild == null)
            return Guild.getErrantGuild();
        else
            return this.owningGuild;
    }

    public void setOwningGuild(Guild owningGuild) {
        this.owningGuild = owningGuild;
    }

    /*
     * Serialization
     */

    public int getFlags() {
        return flags;
    }

    public void setFlags(int flags) {
        this.flags = flags;
    }

    public Zone getParentZone() {
        return parentZone;
    }

    public GuildTag getGuildTag() {
        return guildTag;
    }

    public void setActive(boolean isAc) {

        this.isActive = isAc;
        Building building = BuildingManager.getBuildingFromCache(this.buildingID);
        if (building != null && !this.isActive)
            building.isDeranking.compareAndSet(true, false);
    }

    public boolean validForMine(mbEnums.ResourceType r) {
        if (this.mineType == null)
            return false;
        return this.mineType.validForMine(r, this.isExpansion());
    }

    public void serializeForMineProduction(ByteBufferWriter writer) {
        writer.putInt(this.getObjectType().ordinal());
        writer.putInt(this.getObjectUUID());
        writer.putInt(this.getObjectUUID()); //actually a hash of mine
        //		writer.putInt(0x215C92BB); //this.unknown1);
        writer.putString(this.mineType.name);
        writer.putString(this.zoneName);
        writer.putInt(this.production.resourceHash);
        writer.putInt(this.production.mine_production);
        writer.putInt(this.getModifiedProductionAmount()); //TODO calculate range penalty here
        writer.putInt(3600); //window in seconds
        writer.putInt(this.isExpansion() ? this.mineType.xpacHash : this.mineType.hash);
    }

    @Override
    public void updateDatabase() {
        // TODO Create update logic.
    }

    public int getBuildingID() {
        return buildingID;
    }

    public void setBuildingID(int buildingID) {
        this.buildingID = buildingID;
    }

    public void handleDestroyMine() {

        if (!this.isActive)
            return;

        //remove tags from mine

        this.guildName = "";
        this.nationName = "";
        this.owningGuild = Guild.getErrantGuild();
        this.lastClaimer = null;
        this.wasClaimed = false;

        // Update database

        DbManager.MineQueries.CHANGE_OWNER(this, 0);

        // Update mesh

        Building mineBuilding = BuildingManager.getBuildingFromCache(this.buildingID);

        if (mineBuilding == null) {
            Logger.debug("Null mine building " + this.getObjectUUID() + ". Unable to Load Building with UID " + this.buildingID);
            return;
        }

        mineBuilding.setOwner(null);
        mineBuilding.refresh(false);

        // remove hirelings

        Building building = (Building) getObject(mbEnums.GameObjectType.Building, this.buildingID);
        BuildingManager.cleanupHirelings(building);
    }

    public boolean claimMine(PlayerCharacter claimer) {

        if (claimer == null)
            return false;

        if (!validateClaimer(claimer))
            return false;

        if (!this.isActive) {
            ErrorPopupMsg.sendErrorMsg(claimer, "Can not for to claim inactive mine.");
            return false;
        }

        if (!updateGuildOwner(claimer))
            return false;

        // Successful claim

        this.lastClaimer = claimer;

        return true;
    }

    public boolean depositMineResources() {

        if (this.owningGuild.isEmptyGuild())
            return false;

        if (this.owningGuild.getOwnedCity() == null)
            return false;

        if (this.owningGuild.getOwnedCity().warehouse == null)
            return false;

        return Warehouse.depositFromMine(this, mbEnums.ResourceType.templateLookup.get(this.production.templateID), this.getModifiedProductionAmount(), this.owningGuild.getOwnedCity().warehouse);
    }

    public boolean updateGuildOwner(PlayerCharacter playerCharacter) {

        Building mineBuilding = BuildingManager.getBuildingFromCache(this.buildingID);

        //should never return null, but let's check just in case.

        if (mineBuilding == null) {
            ChatManager.chatSystemError(playerCharacter, "Unable to find mine tower.");
            Logger.debug("Failed to Update Mine with UID " + this.getObjectUUID() + ". Unable to Load Building with UID " + this.buildingID);
            return false;
        }

        if (playerCharacter == null) {
            this.owningGuild = Guild.getErrantGuild();
            this.guildName = "None";
            this.guildTag = GuildTag.ERRANT;
            this.nationName = "None";
            this.nationTag = GuildTag.ERRANT;
            //Update Building.
            mineBuilding.setOwner(null);
            WorldGrid.updateObject(mineBuilding);
            return true;
        }

        Guild guild = playerCharacter.getGuild();

        if (guild.getOwnedCity() == null)
            return false;

        if (!MineQueries.CHANGE_OWNER(this, guild.getObjectUUID())) {
            Logger.debug("Database failed to Change Ownership of Mine with UID " + this.getObjectUUID());
            ChatManager.chatSystemError(playerCharacter, "Failed to claim Mine.");
            return false;
        }

        //update mine.
        this.owningGuild = guild;

        //Update Building.
        PlayerCharacter guildLeader = (PlayerCharacter) Guild.GetGL(this.owningGuild);

        if (guildLeader != null)
            mineBuilding.setOwner(guildLeader);
        WorldGrid.updateObject(mineBuilding);
        return true;
    }

    public boolean isExpansion() {
        return (this.flags & 2) != 0;
    }

    public int getModifiedProductionAmount() {
        //TODO Calculate Distance modifications.

        //calculate base values.
        int baseProduction = this.production.mine_production;
        float baseModValue = this.production.mine_production * .1f;
        float rankModValue = this.production.mine_production * .0143f;
        float totalModded = 0;

        //get Mine Building.
        Building mineBuilding = BuildingManager.getBuilding(this.buildingID);
        if (mineBuilding == null)
            return this.production.mine_production;
        for (AbstractCharacter harvester : mineBuilding.getHirelings().keySet()) {
            totalModded += baseModValue;
            totalModded += rankModValue * harvester.getRank();
        }
        //add base production on top;
        totalModded += baseProduction;
        //skip distance check for expansion.
        if (this.isExpansion())
            return (int) totalModded;

        if (this.owningGuild.isEmptyGuild() == false) {
            if (this.owningGuild.getOwnedCity() != null) {
                float distanceSquared = this.owningGuild.getOwnedCity().getLoc().distanceSquared2D(mineBuilding.getLoc());

                if (distanceSquared > sqr(10000 * 3))
                    totalModded *= .25f;
                else if (distanceSquared > sqr(10000 * 2))
                    totalModded *= .50f;
                else if (distanceSquared > sqr(10000))
                    totalModded *= .75f;
            }
        }
        return (int) totalModded;
    }

}
